/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/def.hpp>
#include <ox/std/error.hpp>
#include <ox/std/string.hpp>

#include <ox/claw/write.hpp>

#include "gfx.hpp"

namespace nostalgia::core {

struct BaseConverter {
	virtual ~BaseConverter() noexcept = default;

	virtual bool matches(const ox::String &srcTypeName, int srcTypeVersion, const ox::String &dstTypeName, int dstTypeVersion) noexcept = 0;

	virtual ox::Error convertRaw(const ox::Buffer &pV1Buff, void *pV2) noexcept = 0;
};

template<typename SrcType, typename DstType>
struct Converter: public BaseConverter{
	bool matches(const ox::String &srcTypeName, int srcTypeVersion, const ox::String &dstTypeName, int dstTypeVersion) noexcept override {
		return srcTypeName == SrcType::TypeName
		    && dstTypeName == DstType::TypeName
		    && srcTypeVersion == SrcType::TypeVersion
		    && dstTypeVersion == DstType::TypeVersion;
	}

	virtual ox::Error convert(SrcType*, DstType*) noexcept {
		return OxError(1);
	}

	ox::Error convertRaw(const ox::Buffer &pV1Buff, void *pV2) noexcept override {
		oxRequireM(src, ox::readClaw<SrcType>(pV1Buff));
		auto dst = static_cast<DstType*>(pV2);
		return convert(&src, dst);
	}
};

ox::Error convert(const ox::Buffer &srcBuffer, const ox::String &dstTypeName, int dstTypeVersion, void *dstObj) noexcept;

template<typename T>
ox::Error convert(const ox::Buffer &srcBuffer, T *dstObj) noexcept {
	return convert((srcBuffer, T::TypeName, T::TypeVersion, &dstObj));
}

template<typename T>
ox::Result<ox::Buffer> convert(const ox::Buffer &srcBuffer, ox::ClawFormat fmt = ox::ClawFormat::Metal) noexcept {
	T dstObj;
	oxReturnError(convert(srcBuffer, T::TypeName, T::TypeVersion, &dstObj));
	return ox::writeClaw(&dstObj, fmt);
}

template<typename T>
ox::Error convert(ox::Buffer *buff) noexcept {
	return convert<T>(*buff).moveTo(buff);

}

}
